import java.awt.*;
import java.util.*;
import java.awt.event.*;
import java.io.*;

/**
 * Server Model - This model keeps track of game state as well
 * as performing game logic.
 * 
 * @author David Couvrette
 * May 30th 2007
 * CSE190L
 * 
 */

public class PongModel implements Serializable, Model {
  // location of ball
  private Point ball;
  
  // height and width of game area
  private transient int width;
  private transient int height;
  
  // velocity of ball
  private transient Point velocity;
  
  // diameter of ball
  private double diameter;
  private Paddle leftPaddle;
  private Paddle rightPaddle;
  private transient PaddleAI ai;
  
  // connect to client
  private transient NetworkServer client;
  
  private boolean inGame;
  
  // list of view that are listening to this model
  private transient java.util.List<Listener> listeners;
  
  private transient javax.swing.Timer timer;
  
  /**
   * Initialize the game state
   * @param width The width of the board
   * @param height The height of the board
   */
  public PongModel(int width, int height) {
    this.width = width;
    this.height = height;
    
    listeners = new java.util.ArrayList<Listener>();
    velocity = new Point(0, 0);
    
    addTimer();
    
    // spawn a thread to wait for incomming connections
    Thread t = new Thread(new PongWaitForConnect(this));
    t.start();
  }
  
  /**
   * Add a listener to current set of listeners
   * @param listener Adds this listener to the model
   */
  public void addListener(Listener listener) {
    listeners.add(listener);
  }
  
  /**
   * Sets a reference to the Network Server
   * @client Reference to NetworkServer 
   */
  public void setClient(NetworkServer client) {
    this.client = client;
  }
  
  /**
   * Notifies the listeners and the connected clients
   * that the state of the model has changed
   */
  private void update() {
    // update network player
    if (client != null) {
      client.sendModel();
    }
    
    for (Listener listener : listeners) {
      listener.update();
    }
  }
  
  /**
   * Starts the game
   */
  public void startGame() {
    if (!inGame) {
      ball = new Point(width / 2, height / 2);
      diameter = 10;
      velocity = new Point(1, 2);
      leftPaddle = new Paddle(height / 2.0, 50.0);
      synchronized (this) {
        rightPaddle = new Paddle(height / 2.0, 50.0);
      }
      //ai = new PaddleAI(this);
      inGame = true;
      timer.start();
    }
  }
  /**
   * Gets the left paddle.
   * @return The left paddle.
   */
  public Paddle getLeftPaddle() {
    return leftPaddle;
  }
  
  /**
   * Gets the right paddle.
   * @return The right paddle.
   */
  public Paddle getRightPaddle() {
    return rightPaddle;
  }
  
  /**
   * Gets the ball.
   * @return The ball.
   */
  public Point getBall() {
    return ball;
  }
  
  /**
   * Gets the diameter of the ball.
   * @return The diameter of the ball.
   */
  public double getDiameter() {
    return diameter;
  }
  
  private void addTimer() {
    timer = new javax.swing.Timer(10, new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        tick();
      }
    });
    timer.setCoalesce(true);
  }
  
  /** 
   * The timer ticked.
   * Update the position of the ball and check where the ball is
   */
  private void tick() {
    if (!inGame) return;
    ball.x += velocity.x;
    ball.y += velocity.y;
    // check bounds
    if (ball.x >= width - 5) {// || ball.x <= 15) {
      // check that the ball is hitting a paddle
      synchronized (this) {
        double diff = Math.abs(ball.y - rightPaddle.getY());
        if (diff < rightPaddle.getHeight() / 2) {
          velocity.x *= -1;
        } else {
          gameOver("Right Loses - Left Wins");
        }
      }
    }
    else if (ball.x <= 15) {
      double diff = Math.abs(ball.y - leftPaddle.getY());
      if (diff < leftPaddle.getHeight() / 2) {
        velocity.x *= -1;
      } else {
        gameOver("Left Loses - Right Wins");
      }
    }
    
    if (ball.y >= height || ball.y <= 0) {
      velocity.y *= -1;
    }
    if (ai != null)
      ai.updatePaddles();
    
    update();
  }
  
  /**
   * Game is over
   * @param winner Name of winner
   */
  private void gameOver(String winner) {
    timer.stop();
    inGame = false;
    System.out.println(winner);
    velocity = null;
    synchronized (this) { 
      rightPaddle = null;
    }
    leftPaddle = null;
    ai = null;
    update();
  }
  
  /**
   * Updates the left paddle position
   * @param y Left paddle position
   */
  public void mouseMoved(double y) {
    // verify that the paddle exists
    if (leftPaddle != null)
      leftPaddle.setY(y);
    update();
  }
  
  /**
   * Updates the paddle position from client
   * @param p This replaces the models current paddle
   */
  public synchronized void updatePaddle(Paddle p) {
    rightPaddle = p;
  }
}